import random
import string


class RandomGenerator:
    @staticmethod
    def nonce(length=40):
        letters = string.ascii_letters + string.digits
        return ''.join(random.choice(letters) for i in range(length))



import hashlib
import hmac
import json
from urllib.parse import quote, urlparse

from pymesomb import mesomb


class Signature:
    @staticmethod
    def sign_request(service, method, url, date, nonce, credentials, headers=None, body=None):
        """
    Method to use to compute signature used in MeSomb request

    :param service: service to use can be payment, wallet ... (the list is provide by MeSomb)
    :param method: HTTP method (GET, POST, PUT, PATCH, DELETE...)
    :param url: the full url of the request with query element https://mesomb.hachther.com/path/to/ressource?highlight=params#url-parsing
    :param date: Datetime of the request
    :param nonce: Unique string generated for each request sent to MeSomb
    :param credentials: dict containing key => value for the credential provided by MeSOmb. {'access' => access_key, 'secret' => secret_key}
    :param headers: Extra HTTP header to use in the signature
    :param body: The dict containing the body you send in your request body
    :return: Authorization to put in the header
    """
        algorithm = mesomb.algorithm
        parse = urlparse(url)
        canonical_query = parse.query

        timestamp = str(int(date.timestamp()))

        # CanonicalHeaders
        if headers is None:
            headers = {}
        headers['host'] = '{}://{}'.format(parse.scheme, parse.netloc)
        headers['x-mesomb-date'] = timestamp
        headers['x-mesomb-nonce'] = nonce
        canonical_headers = '\n'.join(['{}:{}'.format(key.lower(), headers[key].strip()) for key in sorted(headers)])

        if body is None:
            body = {}
        request_params = json.dumps(body, separators=(',', ':'))

        payload_hash = hashlib.sha1(request_params.encode('utf-8')).hexdigest()

        signed_headers = ';'.join(sorted(headers))

        canonical_request = '{}\n{}\n{}\n{}\n{}\n{}'.format(method, quote(parse.path), canonical_query,
                                                            canonical_headers,
                                                            signed_headers, payload_hash)

        scope = '{}/{}/mesomb_request'.format(date.strftime('%Y%m%d'), service)
        string_to_sign = '{}\n{}\n{}\n{}'.format(algorithm, timestamp, scope,
                                                 hashlib.sha1(canonical_request.encode('utf-8')).hexdigest())

        signature = hmac.new(credentials['secret_key'].encode(), string_to_sign.encode(), hashlib.sha1).hexdigest()

        authorization_header = '{} Credential={}/{}, SignedHeaders={}, Signature={}'.format(algorithm,
                                                                                            credentials['access_key'],
                                                                                            scope,
                                                                                            signed_headers, signature)

        return authorization_header



class APIException(Exception):
    """
  Base class for REST framework exceptions.
  Subclasses should provide `.status_code` and `.default_detail` properties.
  """
    default_detail = ''
    default_code = 'error'

    def __init__(self, detail, code):
        self.detail = detail or self.default_detail
        self.code = code or self.default_code

    def __str__(self):
        return str(self.detail)


class ServiceNotFoundException(APIException):
    pass


class PermissionDeniedException(APIException):
    pass


class InvalidClientRequestException(APIException):
    pass


class ServerException(APIException):
    pass



import json
from datetime import datetime

import requests

from pymesomb import mesomb
from pymesomb.exceptions import ServiceNotFoundException, PermissionDeniedException, InvalidClientRequestException, \
    ServerException
from pymesomb.models import TransactionResponse, Application, Transaction
from pymesomb.signature import Signature


def process_client_exception(response):
    status_code = response.status_code
    detail = response.text
    code = None

    if detail.startswith('{'):
        data = json.loads(detail)
        detail = data['detail']
        code = data['code']

    if status_code == 404:
        raise ServiceNotFoundException(detail, code)

    if status_code in [403, 401]:
        raise PermissionDeniedException(detail, code)

    if status_code == 400:
        raise InvalidClientRequestException(detail, code)

    raise ServerException(detail, code)


def build_url(endpoint):
    return '{}/en/api/{}/{}'.format(mesomb.host, mesomb.api_version, endpoint)


class PaymentOperation:
    def __init__(self, application_key, access_key, secret_key):
        self.application_key = application_key
        self.access_key = access_key
        self.secret_key = secret_key

    def get_authorization(self, method, endpoint, date, nonce, headers=None, body=None):
        if headers is None:
            headers = {}

        url = build_url(endpoint)

        credentials = {'access_key': self.access_key, 'secret_key': self.secret_key}

        return Signature.sign_request('payment', method, url, date, nonce, credentials, headers, body)

    def _execute_request(self, method, endpoint, date, nonce='', body=None, mode=None):
        url = build_url(endpoint)

        headers = {
            'x-mesomb-date': str(int(date.timestamp())),
            'x-mesomb-nonce': nonce,
            'X-MeSomb-Application': self.application_key,
        }
        if body:
            body['source'] = 'PyMeSomb/{}'.format(mesomb.version)
        if body and 'trxID' in body:
            headers['X-MeSomb-TrxID'] = body.pop('trxID')

        if mode:
            headers['X-MeSomb-OperationMode'] = mode

        if method == 'POST':
            authorization = self.get_authorization(method, endpoint, date, nonce,
                                                   headers={'content-type': 'application/json'},
                                                   body=body)
        else:
            authorization = self.get_authorization(method, endpoint, date, nonce)

        headers['Authorization'] = authorization

        response = requests.request(method, url, json=body, headers=headers)

        status_code = response.status_code
        if status_code >= 400:
            process_client_exception(response)

        return response.json()

    def make_collect(self, params):
        """
        Collect money a use account
        [Check the documentation here](https://mesomb.hachther.com/en/api/schema/)

        :param params: dict with the below information
            - amount: amount to collect
            - service: payment service with the possible values MTN, ORANGE, AIRTEL
            - payer: account number to collect from
            - date: date of the request
            - nonce: unique string on each request
            - country: 2 letters country code of the service (configured during your service registration in MeSomb)
            - currency: currency of your service depending on your country
            - fees: false if your want MeSomb fees to be computed and included in the amount to collect
            - mode: asynchronous or synchronous
            - conversion: true in case of foreign currently defined if you want to rely on MeSomb to convert the amount in the local currency
            - location: Map containing the location of the customer with the following attributes: town, region and location all string.
            - products: It is array of products. Each product are Map with the following attributes: name string, category string, quantity int and amount float
            - customer: a Map containing information about the customer: phone string, email: string, first_name string, last_name string, address string, town string, region string and country string
            - trxID: if you want to include your transaction ID in the request
            - extra: Map to add some extra attribute depending on the API documentation
        :return: TransactionResponse
        """

        endpoint = 'payment/collect/'

        body = {
            'amount': params['amount'],
            'payer': params['payer'],
            'fees': params.get('fees', True),
            'service': params['service'],
            'country': params.get('country', 'CM'),
            'currency': params.get('currency', 'XAF'),
            'conversion': params.get('conversion', False)
        }
        if 'trxID' in params:
            body['trxID'] = params['trxID']

        if 'extra' in params:
            body.update(params['extra'])

        if 'location' in params:
            body['location'] = params['location']

        if 'customer' in params:
            body['customer'] = params['customer']

        if 'products' in params:
            body['products'] = params['products']

        date = params.get('date', datetime.now())
        nonce = params.get('nonce')

        return TransactionResponse(
            self._execute_request('POST', endpoint, date, nonce, body, params.get('mode', 'synchronous')))

    def make_deposit(self, params):
        """
        Method to make deposit in a receiver mobile account.
        [Check the documentation here](https://mesomb.hachther.com/en/api/schema/)

        :param params: dict with the below information
            - amount: amount to collect
            - service: payment service with the possible values MTN, ORANGE, AIRTEL
            - receiver: account number to depose money
            - date: date of the request
            - nonce: unique string on each request
            - country: 2 letters country code of the service (configured during your service registration in MeSomb)
            - currency: currency of your service depending on your country
            - fees: false if your want MeSomb fees to be computed and included in the amount to collect
            - conversion: true in case of foreign currently defined if you want to rely on MeSomb to convert the amount in the local currency
            - location: Map containing the location of the customer with the following attributes: town, region and location all string.
            - products: It is array of products. Each product are Map with the following attributes: name string, category string, quantity int and amount float
            - customer: a Map containing information about the customer: phone string, email: string, first_name string, last_name string, address string, town string, region string and country string
            - trxID: if you want to include your transaction ID in the request
            - extra: Map to add some extra attribute depending on the API documentation
        :return: TransactionResponse
        """
        endpoint = 'payment/deposit/'
        url = build_url(endpoint)

        body = {
            'amount': params['amount'],
            'receiver': params['receiver'],
            'service': params['service'],
            'country': params.get('country', 'CM'),
            'currency': params.get('currency', 'XAF'),
        }

        if 'extra' in params:
            body.update(params['extra'])

        if 'trxID' in params:
            body['trxID'] = params['trxID']

        if 'location' in params:
            body['location'] = params['location']

        if 'customer' in params:
            body['customer'] = params['customer']

        if 'products' in params:
            body['products'] = params['products']

        date = params.get('date', datetime.now())
        nonce = params['nonce']

        return TransactionResponse(self._execute_request('POST', endpoint, date, nonce, body))

    def update_security(self, field, action, value=None, date=None):
        """
        Update security parameters of your service on MeSomb
        :param field: which security field you want to update (check API doucumentation)
        :param action: SET or UNSET
        :param value: value of the field
        :param date: date of the request
        :return: Application
        """
        endpoint = 'payment/security/'
        body = {'field': field, 'action': action}
        if value:
            body['value'] = value

        if date is None:
            date = datetime.now()

        return Application(self._execute_request('POST', endpoint, date, body=body))

    def get_status(self):
        """
        Get the current status of your service on MeSomb

        :param date: date of the request
        :return: Application
        """
        endpoint = 'payment/status/'

        return Application(self._execute_request('GET', endpoint, datetime.now()))

    def get_transactions(self, ids, source='MESOMB'):
        """
        Get transactions from MeSomb by IDs.

        :param source: source of the transactionID
        :param ids: list of ids
        :return: list of Transaction
        """
        endpoint = 'payment/transactions/?ids={}&source={}'.format(','.join(ids), source)

        return [Transaction(item) for item in self._execute_request('GET', endpoint, datetime.now())]

    def check_transactions(self, ids, source='MESOMB'):
        """
        Get transactions from MeSomb by IDs.

        :param source: source of the transactionID
        :param ids: list of ids
        :return: list of Transaction
        """
        endpoint = 'payment/transactions/check/?ids={}&source={}'.format(','.join(ids), source)
        return [Transaction(item) for item in self._execute_request('GET', endpoint, datetime.now())]



class Location:
    def __init__(self, data):
        self.town = data.get('town')
        self.region = data.get('region')
        self.country = data.get('country')


class Customer:
    def __init__(self, data):
        self.email = data.get('email', None)
        self.phone = data.get('phone', None)
        self.town = data.get('town', None)
        self.region = data.get('region', None)
        self.country = data.get('country', None)
        self.first_name = data.get('first_name', None)
        self.last_name = data.get('last_name', None)
        self.address = data.get('address', None)


class Product:
    def __init__(self, data):
        self.name = data['name']
        self.category = data.get('category', 'category')
        self.quantity = data.get('quantity', None)
        self.amount = data.get('amount', None)


class Transaction:
    def __init__(self, data):
        self.pk = data['pk']
        self.status = data['status']
        self.type = data['type']
        self.amount = data['amount']
        self.fees = data['fees']
        self.b_party = data['b_party']
        if 'message' in data:
            self.message = data['message']
        self.service = data['service']
        if 'reference' in data:
            self.reference = data['reference']
        self.ts = data['ts']
        self.country = data['country']
        self.currency = data['currency']
        if 'fin_trx_id' in data:
            self.fin_trx_id = data['fin_trx_id']
        if 'trxamount' in data:
            self.trxamount = data['trxamount']
        if data.get('location'):
            self.location = Location(data['location'])
        if data.get('customer'):
            self.customer = Customer(data['customer'])
        if data.get('products'):
            self.products = [Product(p) for p in data['products']]
        # self.products = data['products?.map((d: Record < string, any >) = > new Product(d))']


class TransactionResponse:
    def __init__(self, data):
        self.success = data['success']
        self.message = data['message']
        self.redirect = data['redirect']
        self.transaction = Transaction(data['transaction'])
        self.reference = data['reference']
        self.status = data['status']

    def __str__(self):
        return self.data['id']

    def is_operation_success(self):
        return self.success

    def is_transaction_success(self):
        return self.success and self.status == 'SUCCESS'


class ApplicationBalance:
    def __init__(self, data):
        self.country = data.get("country")
        self.currency = data.get("currency")
        self.provider = data.get("provider")
        self.value = data.get("value")
        self.service_name = data.get("service_name")


class Application:
    def __init__(self, data):
        self.key = data['key']
        self.logo = data['logo']
        self.balances = [ApplicationBalance(b) for b in data['balances']]
        self.countries = data['countries']
        self.description = data['description']
        self.isLive = data['is_live']
        self.name = data['name']
        self.security = data['security']
        self.status = data['status']
        self.url = data['url']

    def get_security_field(self, field):
        return self.security.get(field, None)

    def get_balance(self, country=None, service=None):
        balance = 0
        for bal in self.balances:
            if country and bal.country != country:
                continue
            if service and bal.provider != service:
                continue
            balance += bal.value
        return balance
